@@ -3,8 +3,7 @@
 /*               This file is part of the program and library                */
 /*    PaPILO --- Parallel Presolve for Integer and Linear Optimization       */
 /*                                                                           */
-/* Copyright (C) 2020-2022 Konrad-Zuse-Zentrum                               */
-/*                     fuer Informationstechnik Berlin                       */
+/* Copyright (C) 2020-2024 Zuse Institute Berlin (ZIB)                       */
 /*                                                                           */
 /* This program is free software: you can redistribute it and/or modify      */
 /* it under the terms of the GNU Lesser General Public License as published  */
@@ -62,6 +61,7 @@ TEST_CASE( "parallel-row-unchanged", "[presolve]" )
 {
    Num<double> num{};
    double time = 0.0;
+int cause = -1;
    Timer t{ time };
    Message msg{};
    Problem<double> problem = setupProblemWithNoParallelRows();
@@ -76,7 +76,7 @@ TEST_CASE( "parallel-row-unchanged", "[presolve]" )
    Reductions<double> reductions{};
 
    PresolveStatus presolveStatus =
-       presolvingMethod.execute( problem, problemUpdate, num, reductions, t );
+       presolvingMethod.execute( problem, problemUpdate, num, reductions, t, cause );
 
    REQUIRE( presolveStatus == PresolveStatus::kUnchanged );
 }
@@ -86,6 +86,7 @@ TEST_CASE( "parallel-row-two-equations-infeasible-second-row-dominant",
 {
    Num<double> num{};
    double time = 0.0;
+   int cause = -1;
    Timer t{ time };
    Message msg{};
    Problem<double> problem =
@@ -101,16 +102,55 @@ TEST_CASE( "parallel-row-two-equations-infeasible-second-row-dominant",
    Reductions<double> reductions{};
 
    PresolveStatus presolveStatus =
-       presolvingMethod.execute( problem, problemUpdate, num, reductions, t );
+       presolvingMethod.execute( problem, problemUpdate, num, reductions, t, cause );
 
    REQUIRE( presolveStatus == PresolveStatus::kInfeasible );
 }
 
+
+TEST_CASE( "parallel-row-two-identical-equations", "[presolve]" )
+{
+   Num<double> num{};
+   double time = 0.0;
+   int cause = -1;
+   Timer t{ time };
+   Message msg{};
+   Problem<double> problem =
+       setupParallelRowWithTwoParallelEquations( 2.0, -2.0, 1, -1 );
+   Statistics statistics{};
+   PresolveOptions presolveOptions{};
+   PostsolveStorage<double> postsolve =
+       PostsolveStorage<double>( problem, num, presolveOptions );
+   ProblemUpdate<double> problemUpdate( problem, postsolve, statistics,
+                                        presolveOptions, num, msg );
+   problemUpdate.checkChangedActivities();
+   ParallelRowDetection<double> presolvingMethod{};
+   Reductions<double> reductions{};
+
+   PresolveStatus presolveStatus =
+       presolvingMethod.execute( problem, problemUpdate, num, reductions, t, cause );
+
+   REQUIRE( presolveStatus == PresolveStatus::kReduced );
+   REQUIRE( reductions.size() == 4 );
+   REQUIRE( reductions.getReduction( 0 ).col == RowReduction::LOCKED );
+   REQUIRE( reductions.getReduction( 0 ).row == 0 );
+   REQUIRE( reductions.getReduction( 0 ).newval == 0 );
+
+   REQUIRE( reductions.getReduction( 1 ).row == 2 );
+   REQUIRE( reductions.getReduction( 1 ).col == RowReduction::LOCKED );
+   REQUIRE( reductions.getReduction( 1 ).newval == 0 );
+
+   REQUIRE( reductions.getReduction( 3 ).col == RowReduction::REDUNDANT );
+   REQUIRE( reductions.getReduction( 3 ).newval == 0 );
+   REQUIRE( reductions.getReduction( 3 ).row == 2 );
+}
+
 TEST_CASE( "parallel-row-two-equations-infeasible-first-row-dominant",
            "[presolve]" )
 {
    Num<double> num{};
    double time = 0.0;
+int cause = -1;
    Timer t{ time };
    Message msg{};
    Problem<double> problem =
@@ -126,7 +166,7 @@ TEST_CASE( "parallel-row-two-equations-infeasible-first-row-dominant",
    Reductions<double> reductions{};
 
    PresolveStatus presolveStatus =
-       presolvingMethod.execute( problem, problemUpdate, num, reductions, t );
+       presolvingMethod.execute( problem, problemUpdate, num, reductions, t, cause );
 
    REQUIRE( presolveStatus == PresolveStatus::kInfeasible );
 }
@@ -136,6 +176,7 @@ TEST_CASE( "parallel-row-two-equations-feasible-second-row-dominant",
 {
    Num<double> num{};
    double time = 0.0;
+int cause = -1;
    Timer t{ time };
    Message msg{};
    Problem<double> problem =
@@ -151,10 +192,10 @@ TEST_CASE( "parallel-row-two-equations-feasible-second-row-dominant",
    Reductions<double> reductions{};
 
    PresolveStatus presolveStatus =
-       presolvingMethod.execute( problem, problemUpdate, num, reductions, t );
+       presolvingMethod.execute( problem, problemUpdate, num, reductions, t, cause );
 
    REQUIRE( presolveStatus == PresolveStatus::kReduced );
-   REQUIRE( reductions.size() == 3 );
+   REQUIRE( reductions.size() == 4 );
    REQUIRE( reductions.getReduction( 0 ).col == RowReduction::LOCKED );
    REQUIRE( reductions.getReduction( 0 ).row == 2 );
    REQUIRE( reductions.getReduction( 0 ).newval == 0 );
@@ -163,9 +204,9 @@ TEST_CASE( "parallel-row-two-equations-feasible-second-row-dominant",
    REQUIRE( reductions.getReduction( 1 ).col == RowReduction::LOCKED );
    REQUIRE( reductions.getReduction( 1 ).newval == 0 );
 
-   REQUIRE( reductions.getReduction( 2 ).col == RowReduction::REDUNDANT );
-   REQUIRE( reductions.getReduction( 2 ).newval == 0 );
-   REQUIRE( reductions.getReduction( 2 ).row == 0 );
+   REQUIRE( reductions.getReduction( 3 ).col == RowReduction::REDUNDANT );
+   REQUIRE( reductions.getReduction( 3 ).newval == 0 );
+   REQUIRE( reductions.getReduction( 3 ).row == 0 );
 }
 
 TEST_CASE( "parallel-row-two-equations-feasible-first-row-dominant",
@@ -173,6 +214,7 @@ TEST_CASE( "parallel-row-two-equations-feasible-first-row-dominant",
 {
    Num<double> num{};
    double time = 0.0;
+   int cause = -1;
    Timer t{ time };
    Message msg{};
    Problem<double> problem =
@@ -188,10 +230,10 @@ TEST_CASE( "parallel-row-two-equations-feasible-first-row-dominant",
    Reductions<double> reductions{};
 
    PresolveStatus presolveStatus =
-       presolvingMethod.execute( problem, problemUpdate, num, reductions, t );
+       presolvingMethod.execute( problem, problemUpdate, num, reductions, t, cause );
 
    REQUIRE( presolveStatus == PresolveStatus::kReduced );
-   REQUIRE( reductions.size() == 3 );
+   REQUIRE( reductions.size() == 4 );
    REQUIRE( reductions.getReduction( 0 ).col == RowReduction::LOCKED );
    REQUIRE( reductions.getReduction( 0 ).row == 0 );
    REQUIRE( reductions.getReduction( 0 ).newval == 0 );
@@ -200,9 +242,9 @@ TEST_CASE( "parallel-row-two-equations-feasible-first-row-dominant",
    REQUIRE( reductions.getReduction( 1 ).col == RowReduction::LOCKED );
    REQUIRE( reductions.getReduction( 1 ).newval == 0 );
 
-   REQUIRE( reductions.getReduction( 2 ).col == RowReduction::REDUNDANT );
-   REQUIRE( reductions.getReduction( 2 ).newval == 0 );
-   REQUIRE( reductions.getReduction( 2 ).row == 2 );
+   REQUIRE( reductions.getReduction( 3 ).col == RowReduction::REDUNDANT );
+   REQUIRE( reductions.getReduction( 3 ).newval == 0 );
+   REQUIRE( reductions.getReduction( 3 ).row == 2 );
 }
 
 TEST_CASE( "parallel-row-two-inequalities-redundant-row-second-row-dominant",
@@ -210,6 +252,7 @@ TEST_CASE( "parallel-row-two-inequalities-redundant-row-second-row-dominant",
 {
    Num<double> num{};
    double time = 0.0;
+int cause = -1;
    Timer t{ time };
    Message msg{};
    Problem<double> problem =
@@ -225,10 +268,10 @@ TEST_CASE( "parallel-row-two-inequalities-redundant-row-second-row-dominant",
    Reductions<double> reductions{};
 
    PresolveStatus presolveStatus =
-       presolvingMethod.execute( problem, problemUpdate, num, reductions, t );
+       presolvingMethod.execute( problem, problemUpdate, num, reductions, t, cause );
 
    REQUIRE( presolveStatus == PresolveStatus::kReduced );
-   REQUIRE( reductions.size() == 3 );
+   REQUIRE( reductions.size() == 4 );
    REQUIRE( reductions.getReduction( 0 ).col == RowReduction::LOCKED );
    REQUIRE( reductions.getReduction( 0 ).row == 2 );
    REQUIRE( reductions.getReduction( 0 ).newval == 0 );
@@ -237,9 +280,9 @@ TEST_CASE( "parallel-row-two-inequalities-redundant-row-second-row-dominant",
    REQUIRE( reductions.getReduction( 1 ).col == RowReduction::LOCKED );
    REQUIRE( reductions.getReduction( 1 ).newval == 0 );
 
-   REQUIRE( reductions.getReduction( 2 ).col == RowReduction::REDUNDANT );
-   REQUIRE( reductions.getReduction( 2 ).newval == 0 );
-   REQUIRE( reductions.getReduction( 2 ).row == 0 );
+   REQUIRE( reductions.getReduction( 3 ).col == RowReduction::REDUNDANT );
+   REQUIRE( reductions.getReduction( 3 ).newval == 0 );
+   REQUIRE( reductions.getReduction( 3 ).row == 0 );
 }
 
 TEST_CASE( "parallel-row-two-inequalities-redundant-row-first-row-dominant",
@@ -247,6 +290,7 @@ TEST_CASE( "parallel-row-two-inequalities-redundant-row-first-row-dominant",
 {
    Num<double> num{};
    double time = 0.0;
+int cause = -1;
    Timer t{ time };
    Message msg{};
    Problem<double> problem =
@@ -262,10 +306,10 @@ TEST_CASE( "parallel-row-two-inequalities-redundant-row-first-row-dominant",
    Reductions<double> reductions{};
 
    PresolveStatus presolveStatus =
-       presolvingMethod.execute( problem, problemUpdate, num, reductions, t );
+       presolvingMethod.execute( problem, problemUpdate, num, reductions, t, cause );
 
    REQUIRE( presolveStatus == PresolveStatus::kReduced );
-   REQUIRE( reductions.size() == 3 );
+   REQUIRE( reductions.size() == 4 );
    REQUIRE( reductions.getReduction( 0 ).col == RowReduction::LOCKED );
    REQUIRE( reductions.getReduction( 0 ).row == 0 );
    REQUIRE( reductions.getReduction( 0 ).newval == 0 );
@@ -274,9 +318,9 @@ TEST_CASE( "parallel-row-two-inequalities-redundant-row-first-row-dominant",
    REQUIRE( reductions.getReduction( 1 ).col == RowReduction::LOCKED );
    REQUIRE( reductions.getReduction( 1 ).newval == 0 );
 
-   REQUIRE( reductions.getReduction( 2 ).col == RowReduction::REDUNDANT );
-   REQUIRE( reductions.getReduction( 2 ).newval == 0 );
-   REQUIRE( reductions.getReduction( 2 ).row == 2 );
+   REQUIRE( reductions.getReduction( 3 ).col == RowReduction::REDUNDANT );
+   REQUIRE( reductions.getReduction( 3 ).newval == 0 );
+   REQUIRE( reductions.getReduction( 3 ).row == 2 );
 }
 
 TEST_CASE(
@@ -285,6 +329,7 @@ TEST_CASE(
 {
    Num<double> num{};
    double time = 0.0;
+int cause = -1;
    Timer t{ time };
    Message msg{};
    Problem<double> problem =
@@ -300,10 +345,10 @@ TEST_CASE(
    Reductions<double> reductions{};
 
    PresolveStatus presolveStatus =
-       presolvingMethod.execute( problem, problemUpdate, num, reductions, t );
+       presolvingMethod.execute( problem, problemUpdate, num, reductions, t, cause );
 
    REQUIRE( presolveStatus == PresolveStatus::kReduced );
-   REQUIRE( reductions.size() == 5 );
+   REQUIRE( reductions.size() == 6 );
    REQUIRE( reductions.getReduction( 0 ).col == RowReduction::LOCKED );
    REQUIRE( reductions.getReduction( 0 ).row == 2 );
    REQUIRE( reductions.getReduction( 0 ).newval == 0 );
@@ -312,19 +357,19 @@ TEST_CASE(
    REQUIRE( reductions.getReduction( 1 ).col == RowReduction::LOCKED );
    REQUIRE( reductions.getReduction( 1 ).newval == 0 );
 
-   REQUIRE( reductions.getReduction( 2 ).col ==
-            RowReduction::REASON_FOR_LESS_RESTRICTIVE_BOUND_CHANGE );
-   REQUIRE( reductions.getReduction( 2 ).newval == 2 );
-   REQUIRE( reductions.getReduction( 2 ).row == 0 );
-
    REQUIRE( reductions.getReduction( 3 ).col ==
-            RowReduction::LHS_LESS_RESTRICTIVE );
-   REQUIRE( reductions.getReduction( 3 ).newval == 0 );
-   REQUIRE( reductions.getReduction( 3 ).row == 2 );
+            RowReduction::REASON_FOR_LESS_RESTRICTIVE_BOUND_CHANGE );
+   REQUIRE( reductions.getReduction( 3 ).newval == 2 );
+   REQUIRE( reductions.getReduction( 3 ).row == 0 );
 
-   REQUIRE( reductions.getReduction( 4 ).col == RowReduction::REDUNDANT );
+   REQUIRE( reductions.getReduction( 4 ).col ==
+            RowReduction::LHS_LESS_RESTRICTIVE );
    REQUIRE( reductions.getReduction( 4 ).newval == 0 );
-   REQUIRE( reductions.getReduction( 4 ).row == 0 );
+   REQUIRE( reductions.getReduction( 4 ).row == 2 );
+
+   REQUIRE( reductions.getReduction( 5 ).col == RowReduction::REDUNDANT );
+   REQUIRE( reductions.getReduction( 5 ).newval == 0 );
+   REQUIRE( reductions.getReduction( 5 ).row == 0 );
 }
 
 TEST_CASE(
@@ -333,6 +378,7 @@ TEST_CASE(
 {
    Num<double> num{};
    double time = 0.0;
+int cause = -1;
    Timer t{ time };
    Message msg{};
    Problem<double> problem =
@@ -348,10 +394,10 @@ TEST_CASE(
    Reductions<double> reductions{};
 
    PresolveStatus presolveStatus =
-       presolvingMethod.execute( problem, problemUpdate, num, reductions, t );
+       presolvingMethod.execute( problem, problemUpdate, num, reductions, t, cause );
 
    REQUIRE( presolveStatus == PresolveStatus::kReduced );
-   REQUIRE( reductions.size() == 5 );
+   REQUIRE( reductions.size() == 6 );
    REQUIRE( reductions.getReduction( 0 ).col == RowReduction::LOCKED );
    REQUIRE( reductions.getReduction( 0 ).row == 0 );
    REQUIRE( reductions.getReduction( 0 ).newval == 0 );
@@ -360,19 +406,19 @@ TEST_CASE(
    REQUIRE( reductions.getReduction( 1 ).col == RowReduction::LOCKED );
    REQUIRE( reductions.getReduction( 1 ).newval == 0 );
 
-   REQUIRE( reductions.getReduction( 2 ).col ==
-            RowReduction::REASON_FOR_LESS_RESTRICTIVE_BOUND_CHANGE );
-   REQUIRE( reductions.getReduction( 2 ).newval == 0 );
-   REQUIRE( reductions.getReduction( 2 ).row == 2 );
-
    REQUIRE( reductions.getReduction( 3 ).col ==
-            RowReduction::LHS_LESS_RESTRICTIVE );
-   REQUIRE( reductions.getReduction( 3 ).newval == -3 );
-   REQUIRE( reductions.getReduction( 3 ).row == 0 );
+            RowReduction::REASON_FOR_LESS_RESTRICTIVE_BOUND_CHANGE );
+   REQUIRE( reductions.getReduction( 3 ).newval == 0 );
+   REQUIRE( reductions.getReduction( 3 ).row == 2 );
 
-   REQUIRE( reductions.getReduction( 4 ).col == RowReduction::REDUNDANT );
-   REQUIRE( reductions.getReduction( 4 ).newval == 0 );
-   REQUIRE( reductions.getReduction( 4 ).row == 2 );
+   REQUIRE( reductions.getReduction( 4 ).col ==
+            RowReduction::LHS_LESS_RESTRICTIVE );
+   REQUIRE( reductions.getReduction( 4 ).newval == -3 );
+   REQUIRE( reductions.getReduction( 4 ).row == 0 );
+
+   REQUIRE( reductions.getReduction( 5 ).col == RowReduction::REDUNDANT );
+   REQUIRE( reductions.getReduction( 5 ).newval == 0 );
+   REQUIRE( reductions.getReduction( 5 ).row == 2 );
 }
 
 TEST_CASE(
@@ -381,6 +427,7 @@ TEST_CASE(
 {
    Num<double> num{};
    double time = 0.0;
+int cause = -1;
    Timer t{ time };
    Message msg{};
    Problem<double> problem =
@@ -396,10 +443,10 @@ TEST_CASE(
    Reductions<double> reductions{};
 
    PresolveStatus presolveStatus =
-       presolvingMethod.execute( problem, problemUpdate, num, reductions, t );
+       presolvingMethod.execute( problem, problemUpdate, num, reductions, t, cause );
 
    REQUIRE( presolveStatus == PresolveStatus::kReduced );
-   REQUIRE( reductions.size() == 5 );
+   REQUIRE( reductions.size() == 6 );
    REQUIRE( reductions.getReduction( 0 ).col == RowReduction::LOCKED );
    REQUIRE( reductions.getReduction( 0 ).row == 2 );
    REQUIRE( reductions.getReduction( 0 ).newval == 0 );
@@ -408,19 +455,19 @@ TEST_CASE(
    REQUIRE( reductions.getReduction( 1 ).col == RowReduction::LOCKED );
    REQUIRE( reductions.getReduction( 1 ).newval == 0 );
 
-   REQUIRE( reductions.getReduction( 2 ).col ==
-            RowReduction::REASON_FOR_LESS_RESTRICTIVE_BOUND_CHANGE );
-   REQUIRE( reductions.getReduction( 2 ).newval == 2 );
-   REQUIRE( reductions.getReduction( 2 ).row == 0 );
-
    REQUIRE( reductions.getReduction( 3 ).col ==
-            RowReduction::RHS_LESS_RESTRICTIVE );
-   REQUIRE( reductions.getReduction( 3 ).newval == 3 );
-   REQUIRE( reductions.getReduction( 3 ).row == 2 );
+            RowReduction::REASON_FOR_LESS_RESTRICTIVE_BOUND_CHANGE );
+   REQUIRE( reductions.getReduction( 3 ).newval == 2 );
+   REQUIRE( reductions.getReduction( 3 ).row == 0 );
 
-   REQUIRE( reductions.getReduction( 4 ).col == RowReduction::REDUNDANT );
-   REQUIRE( reductions.getReduction( 4 ).newval == 0 );
-   REQUIRE( reductions.getReduction( 4 ).row == 0 );
+   REQUIRE( reductions.getReduction( 4 ).col ==
+            RowReduction::RHS_LESS_RESTRICTIVE );
+   REQUIRE( reductions.getReduction( 4 ).newval == 3 );
+   REQUIRE( reductions.getReduction( 4 ).row == 2 );
+
+   REQUIRE( reductions.getReduction( 5 ).col == RowReduction::REDUNDANT );
+   REQUIRE( reductions.getReduction( 5 ).newval == 0 );
+   REQUIRE( reductions.getReduction( 5 ).row == 0 );
 }
 
 TEST_CASE(
@@ -429,6 +476,7 @@ TEST_CASE(
 {
    Num<double> num{};
    double time = 0.0;
+int cause = -1;
    Timer t{ time };
    Message msg{};
    Problem<double> problem =
@@ -444,10 +492,10 @@ TEST_CASE(
    Reductions<double> reductions{};
 
    PresolveStatus presolveStatus =
-       presolvingMethod.execute( problem, problemUpdate, num, reductions, t );
+       presolvingMethod.execute( problem, problemUpdate, num, reductions, t, cause );
 
    REQUIRE( presolveStatus == PresolveStatus::kReduced );
-   REQUIRE( reductions.size() == 5 );
+   REQUIRE( reductions.size() == 6 );
    REQUIRE( reductions.getReduction( 0 ).col == RowReduction::LOCKED );
    REQUIRE( reductions.getReduction( 0 ).row == 0 );
    REQUIRE( reductions.getReduction( 0 ).newval == 0 );
@@ -456,19 +504,19 @@ TEST_CASE(
    REQUIRE( reductions.getReduction( 1 ).col == RowReduction::LOCKED );
    REQUIRE( reductions.getReduction( 1 ).newval == 0 );
 
-   REQUIRE( reductions.getReduction( 2 ).col ==
-            RowReduction::REASON_FOR_LESS_RESTRICTIVE_BOUND_CHANGE );
-   REQUIRE( reductions.getReduction( 2 ).newval == 0 );
-   REQUIRE( reductions.getReduction( 2 ).row == 2 );
-
    REQUIRE( reductions.getReduction( 3 ).col ==
-            RowReduction::RHS_LESS_RESTRICTIVE );
-   REQUIRE( reductions.getReduction( 3 ).newval == 3 );
-   REQUIRE( reductions.getReduction( 3 ).row == 0 );
+            RowReduction::REASON_FOR_LESS_RESTRICTIVE_BOUND_CHANGE );
+   REQUIRE( reductions.getReduction( 3 ).newval == 0 );
+   REQUIRE( reductions.getReduction( 3 ).row == 2 );
 
-   REQUIRE( reductions.getReduction( 4 ).col == RowReduction::REDUNDANT );
-   REQUIRE( reductions.getReduction( 4 ).newval == 0 );
-   REQUIRE( reductions.getReduction( 4 ).row == 2 );
+   REQUIRE( reductions.getReduction( 4 ).col ==
+            RowReduction::RHS_LESS_RESTRICTIVE );
+   REQUIRE( reductions.getReduction( 4 ).newval == 3 );
+   REQUIRE( reductions.getReduction( 4 ).row == 0 );
+
+   REQUIRE( reductions.getReduction( 5 ).col == RowReduction::REDUNDANT );
+   REQUIRE( reductions.getReduction( 5 ).newval == 0 );
+   REQUIRE( reductions.getReduction( 5 ).row == 2 );
 }
 
 TEST_CASE( "parallel-row-two-inequalities-infeasible-first-row-dominant",
@@ -476,6 +524,7 @@ TEST_CASE( "parallel-row-two-inequalities-infeasible-first-row-dominant",
 {
    Num<double> num{};
    double time = 0.0;
+int cause = -1;
    Timer t{ time };
    Message msg{};
    Problem<double> problem =
@@ -491,7 +540,7 @@ TEST_CASE( "parallel-row-two-inequalities-infeasible-first-row-dominant",
    Reductions<double> reductions{};
 
    PresolveStatus presolveStatus =
-       presolvingMethod.execute( problem, problemUpdate, num, reductions, t );
+       presolvingMethod.execute( problem, problemUpdate, num, reductions, t, cause );
 
    REQUIRE( presolveStatus == PresolveStatus::kInfeasible );
 }
@@ -501,6 +550,7 @@ TEST_CASE( "parallel-row-two-inequalities-infeasible-second-row-dominant",
 {
    Num<double> num{};
    double time = 0.0;
+int cause = -1;
    Timer t{ time };
    Message msg{};
    Problem<double> problem =
@@ -516,7 +566,7 @@ TEST_CASE( "parallel-row-two-inequalities-infeasible-second-row-dominant",
    Reductions<double> reductions{};
 
    PresolveStatus presolveStatus =
-       presolvingMethod.execute( problem, problemUpdate, num, reductions, t );
+       presolvingMethod.execute( problem, problemUpdate, num, reductions, t, cause );
 
    REQUIRE( presolveStatus == PresolveStatus::kInfeasible );
 }
@@ -527,6 +577,7 @@ TEST_CASE( "parallel-row-two-inequalities-tighten-upper-bound-first-row-neg"
 {
    Num<double> num{};
    double time = 0.0;
+int cause = -1;
    Timer t{ time };
    Message msg{};
    Problem<double> problem =
@@ -542,10 +593,10 @@ TEST_CASE( "parallel-row-two-inequalities-tighten-upper-bound-first-row-neg"
    Reductions<double> reductions{};
 
    PresolveStatus presolveStatus =
-       presolvingMethod.execute( problem, problemUpdate, num, reductions, t );
+       presolvingMethod.execute( problem, problemUpdate, num, reductions, t, cause );
 
    REQUIRE( presolveStatus == PresolveStatus::kReduced );
-   REQUIRE( reductions.size() == 5 );
+   REQUIRE( reductions.size() == 6 );
    REQUIRE( reductions.getReduction( 0 ).col == RowReduction::LOCKED );
    REQUIRE( reductions.getReduction( 0 ).row == 0 );
    REQUIRE( reductions.getReduction( 0 ).newval == 0 );
@@ -554,25 +605,26 @@ TEST_CASE( "parallel-row-two-inequalities-tighten-upper-bound-first-row-neg"
    REQUIRE( reductions.getReduction( 1 ).col == RowReduction::LOCKED );
    REQUIRE( reductions.getReduction( 1 ).newval == 0 );
 
-   REQUIRE( reductions.getReduction( 2 ).col ==
-            RowReduction::REASON_FOR_LESS_RESTRICTIVE_BOUND_CHANGE );
-   REQUIRE( reductions.getReduction( 2 ).newval == 0 );
-   REQUIRE( reductions.getReduction( 2 ).row == 2 );
-
    REQUIRE( reductions.getReduction( 3 ).col ==
-            RowReduction::RHS_LESS_RESTRICTIVE );
-   REQUIRE( reductions.getReduction( 3 ).newval == 3 );
-   REQUIRE( reductions.getReduction( 3 ).row == 0 );
+            RowReduction::REASON_FOR_LESS_RESTRICTIVE_BOUND_CHANGE );
+   REQUIRE( reductions.getReduction( 3 ).newval == 0 );
+   REQUIRE( reductions.getReduction( 3 ).row == 2 );
 
-   REQUIRE( reductions.getReduction( 4 ).col == RowReduction::REDUNDANT );
-   REQUIRE( reductions.getReduction( 4 ).newval == 0 );
-   REQUIRE( reductions.getReduction( 4 ).row == 2 );
+   REQUIRE( reductions.getReduction( 4 ).col ==
+            RowReduction::RHS_LESS_RESTRICTIVE );
+   REQUIRE( reductions.getReduction( 4 ).newval == 3 );
+   REQUIRE( reductions.getReduction( 4 ).row == 0 );
+
+   REQUIRE( reductions.getReduction( 5 ).col == RowReduction::REDUNDANT );
+   REQUIRE( reductions.getReduction( 5 ).newval == 0 );
+   REQUIRE( reductions.getReduction( 5 ).row == 2 );
 }
 
 TEST_CASE( "parallel-row-overwrite-inf-first-row-rhs-inf", "[presolve]" )
 {
    Num<double> num{};
    double time = 0.0;
+int cause = -1;
    Timer t{ time };
    Message msg{};
    Problem<double> problem =
@@ -588,10 +640,10 @@ TEST_CASE( "parallel-row-overwrite-inf-first-row-rhs-inf", "[presolve]" )
    Reductions<double> reductions{};
 
    PresolveStatus presolveStatus =
-       presolvingMethod.execute( problem, problemUpdate, num, reductions, t );
+       presolvingMethod.execute( problem, problemUpdate, num, reductions, t, cause );
 
    REQUIRE( presolveStatus == PresolveStatus::kReduced );
-   REQUIRE( reductions.size() == 5 );
+   REQUIRE( reductions.size() == 6 );
    REQUIRE( reductions.getReduction( 0 ).col == RowReduction::LOCKED );
    REQUIRE( reductions.getReduction( 0 ).row == 0 );
    REQUIRE( reductions.getReduction( 0 ).newval == 0 );
@@ -600,25 +652,26 @@ TEST_CASE( "parallel-row-overwrite-inf-first-row-rhs-inf", "[presolve]" )
    REQUIRE( reductions.getReduction( 1 ).col == RowReduction::LOCKED );
    REQUIRE( reductions.getReduction( 1 ).newval == 0 );
 
-   REQUIRE( reductions.getReduction( 2 ).col ==
-            RowReduction::REASON_FOR_LESS_RESTRICTIVE_BOUND_CHANGE );
-   REQUIRE( reductions.getReduction( 2 ).newval == 0 );
-   REQUIRE( reductions.getReduction( 2 ).row == 2 );
-
    REQUIRE( reductions.getReduction( 3 ).col ==
-            RowReduction::RHS_LESS_RESTRICTIVE );
-   REQUIRE( reductions.getReduction( 3 ).newval == 2 );
-   REQUIRE( reductions.getReduction( 3 ).row == 0 );
+            RowReduction::REASON_FOR_LESS_RESTRICTIVE_BOUND_CHANGE );
+   REQUIRE( reductions.getReduction( 3 ).newval == 0 );
+   REQUIRE( reductions.getReduction( 3 ).row == 2 );
 
-   REQUIRE( reductions.getReduction( 4 ).col == RowReduction::REDUNDANT );
-   REQUIRE( reductions.getReduction( 4 ).newval == 0 );
-   REQUIRE( reductions.getReduction( 4 ).row == 2 );
+   REQUIRE( reductions.getReduction( 4 ).col ==
+            RowReduction::RHS_LESS_RESTRICTIVE );
+   REQUIRE( reductions.getReduction( 4 ).newval == 2 );
+   REQUIRE( reductions.getReduction( 4 ).row == 0 );
+
+   REQUIRE( reductions.getReduction( 5 ).col == RowReduction::REDUNDANT );
+   REQUIRE( reductions.getReduction( 5 ).newval == 0 );
+   REQUIRE( reductions.getReduction( 5 ).row == 2 );
 }
 
 TEST_CASE( "parallel-row-overwrite-inf-first-row-lhs-inf", "[presolve]" )
 {
    Num<double> num{};
    double time = 0.0;
+int cause = -1;
    Timer t{ time };
    Message msg{};
    Problem<double> problem =
@@ -634,10 +687,10 @@ TEST_CASE( "parallel-row-overwrite-inf-first-row-lhs-inf", "[presolve]" )
    Reductions<double> reductions{};
 
    PresolveStatus presolveStatus =
-       presolvingMethod.execute( problem, problemUpdate, num, reductions, t );
+       presolvingMethod.execute( problem, problemUpdate, num, reductions, t, cause );
 
    REQUIRE( presolveStatus == PresolveStatus::kReduced );
-   REQUIRE( reductions.size() == 5 );
+   REQUIRE( reductions.size() == 6 );
    REQUIRE( reductions.getReduction( 0 ).col == RowReduction::LOCKED );
    REQUIRE( reductions.getReduction( 0 ).row == 0 );
    REQUIRE( reductions.getReduction( 0 ).newval == 0 );
@@ -646,19 +699,19 @@ TEST_CASE( "parallel-row-overwrite-inf-first-row-lhs-inf", "[presolve]" )
    REQUIRE( reductions.getReduction( 1 ).col == RowReduction::LOCKED );
    REQUIRE( reductions.getReduction( 1 ).newval == 0 );
 
-   REQUIRE( reductions.getReduction( 2 ).col ==
-            RowReduction::REASON_FOR_LESS_RESTRICTIVE_BOUND_CHANGE );
-   REQUIRE( reductions.getReduction( 2 ).newval == 0 );
-   REQUIRE( reductions.getReduction( 2 ).row == 2 );
-
    REQUIRE( reductions.getReduction( 3 ).col ==
-            RowReduction::LHS_LESS_RESTRICTIVE );
-   REQUIRE( reductions.getReduction( 3 ).newval == 2 );
-   REQUIRE( reductions.getReduction( 3 ).row == 0 );
+            RowReduction::REASON_FOR_LESS_RESTRICTIVE_BOUND_CHANGE );
+   REQUIRE( reductions.getReduction( 3 ).newval == 0 );
+   REQUIRE( reductions.getReduction( 3 ).row == 2 );
 
-   REQUIRE( reductions.getReduction( 4 ).col == RowReduction::REDUNDANT );
-   REQUIRE( reductions.getReduction( 4 ).newval == 0 );
-   REQUIRE( reductions.getReduction( 4 ).row == 2 );
+   REQUIRE( reductions.getReduction( 4 ).col ==
+            RowReduction::LHS_LESS_RESTRICTIVE );
+   REQUIRE( reductions.getReduction( 4 ).newval == 2 );
+   REQUIRE( reductions.getReduction( 4 ).row == 0 );
+
+   REQUIRE( reductions.getReduction( 5 ).col == RowReduction::REDUNDANT );
+   REQUIRE( reductions.getReduction( 5 ).newval == 0 );
+   REQUIRE( reductions.getReduction( 5 ).row == 2 );
 }
 
 TEST_CASE( "parallel-row-overwrite-inf-first-row-lhs-inf-neg-factor",
@@ -666,6 +719,7 @@ TEST_CASE( "parallel-row-overwrite-inf-first-row-lhs-inf-neg-factor",
 {
    Num<double> num{};
    double time = 0.0;
+int cause = -1;
    Timer t{ time };
    Message msg{};
    Problem<double> problem =
@@ -681,10 +735,10 @@ TEST_CASE( "parallel-row-overwrite-inf-first-row-lhs-inf-neg-factor",
    Reductions<double> reductions{};
 
    PresolveStatus presolveStatus =
-       presolvingMethod.execute( problem, problemUpdate, num, reductions, t );
+       presolvingMethod.execute( problem, problemUpdate, num, reductions, t, cause );
 
    REQUIRE( presolveStatus == PresolveStatus::kReduced );
-   REQUIRE( reductions.size() == 5 );
+   REQUIRE( reductions.size() == 6 );
    REQUIRE( reductions.getReduction( 0 ).col == RowReduction::LOCKED );
    REQUIRE( reductions.getReduction( 0 ).row == 0 );
    REQUIRE( reductions.getReduction( 0 ).newval == 0 );
@@ -693,25 +747,26 @@ TEST_CASE( "parallel-row-overwrite-inf-first-row-lhs-inf-neg-factor",
    REQUIRE( reductions.getReduction( 1 ).col == RowReduction::LOCKED );
    REQUIRE( reductions.getReduction( 1 ).newval == 0 );
 
-   REQUIRE( reductions.getReduction( 2 ).col ==
-            RowReduction::REASON_FOR_LESS_RESTRICTIVE_BOUND_CHANGE );
-   REQUIRE( reductions.getReduction( 2 ).newval == 0 );
-   REQUIRE( reductions.getReduction( 2 ).row == 2 );
-
    REQUIRE( reductions.getReduction( 3 ).col ==
-            RowReduction::RHS_LESS_RESTRICTIVE );
-   REQUIRE( reductions.getReduction( 3 ).newval == 1 );
-   REQUIRE( reductions.getReduction( 3 ).row == 0 );
+            RowReduction::REASON_FOR_LESS_RESTRICTIVE_BOUND_CHANGE );
+   REQUIRE( reductions.getReduction( 3 ).newval == 0 );
+   REQUIRE( reductions.getReduction( 3 ).row == 2 );
 
-   REQUIRE( reductions.getReduction( 4 ).col == RowReduction::REDUNDANT );
-   REQUIRE( reductions.getReduction( 4 ).newval == 0 );
-   REQUIRE( reductions.getReduction( 4 ).row == 2 );
+   REQUIRE( reductions.getReduction( 4 ).col ==
+            RowReduction::RHS_LESS_RESTRICTIVE );
+   REQUIRE( reductions.getReduction( 4 ).newval == 1 );
+   REQUIRE( reductions.getReduction( 4 ).row == 0 );
+
+   REQUIRE( reductions.getReduction( 5 ).col == RowReduction::REDUNDANT );
+   REQUIRE( reductions.getReduction( 5 ).newval == 0 );
+   REQUIRE( reductions.getReduction( 5 ).row == 2 );
 }
 
 TEST_CASE( "parallel-row-mixed-infeasible-first-row-equation", "[presolve]" )
 {
    Num<double> num{};
    double time = 0.0;
+int cause = -1;
    Timer t{ time };
    Message msg{};
    Problem<double> problem =
@@ -727,7 +782,7 @@ TEST_CASE( "parallel-row-mixed-infeasible-first-row-equation", "[presolve]" )
    Reductions<double> reductions{};
 
    PresolveStatus presolveStatus =
-       presolvingMethod.execute( problem, problemUpdate, num, reductions, t );
+       presolvingMethod.execute( problem, problemUpdate, num, reductions, t, cause );
 
    REQUIRE( presolveStatus == PresolveStatus::kInfeasible );
 }
@@ -736,6 +791,7 @@ TEST_CASE( "parallel-row-mixed-second-row-equation", "[presolve]" )
 {
    Num<double> num{};
    double time = 0.0;
+int cause = -1;
    Timer t{ time };
    Message msg{};
    Problem<double> problem =
@@ -751,10 +807,10 @@ TEST_CASE( "parallel-row-mixed-second-row-equation", "[presolve]" )
    Reductions<double> reductions{};
 
    PresolveStatus presolveStatus =
-       presolvingMethod.execute( problem, problemUpdate, num, reductions, t );
+       presolvingMethod.execute( problem, problemUpdate, num, reductions, t, cause );
 
    REQUIRE( presolveStatus == PresolveStatus::kReduced );
-   REQUIRE( reductions.size() == 3 );
+   REQUIRE( reductions.size() == 4 );
    REQUIRE( reductions.getReduction( 0 ).col == RowReduction::LOCKED );
    REQUIRE( reductions.getReduction( 0 ).row == 2 );
    REQUIRE( reductions.getReduction( 0 ).newval == 0 );
@@ -763,15 +819,16 @@ TEST_CASE( "parallel-row-mixed-second-row-equation", "[presolve]" )
    REQUIRE( reductions.getReduction( 1 ).col == RowReduction::LOCKED );
    REQUIRE( reductions.getReduction( 1 ).newval == 0 );
 
-   REQUIRE( reductions.getReduction( 2 ).col == RowReduction::REDUNDANT );
-   REQUIRE( reductions.getReduction( 2 ).newval == 0 );
-   REQUIRE( reductions.getReduction( 2 ).row == 0 );
+   REQUIRE( reductions.getReduction( 3 ).col == RowReduction::REDUNDANT );
+   REQUIRE( reductions.getReduction( 3 ).newval == 0 );
+   REQUIRE( reductions.getReduction( 3 ).row == 0 );
 }
 
 TEST_CASE( "parallel-row-mixed-infeasible-second-row-equation", "[presolve]" )
 {
    Num<double> num{};
    double time = 0.0;
+int cause = -1;
    Timer t{ time };
    Message msg{};
    Problem<double> problem =
@@ -787,7 +844,7 @@ TEST_CASE( "parallel-row-mixed-infeasible-second-row-equation", "[presolve]" )
    Reductions<double> reductions{};
 
    PresolveStatus presolveStatus =
-       presolvingMethod.execute( problem, problemUpdate, num, reductions, t );
+       presolvingMethod.execute( problem, problemUpdate, num, reductions, t, cause );
 
    REQUIRE( presolveStatus == PresolveStatus::kInfeasible );
 }
@@ -796,6 +853,7 @@ TEST_CASE( "parallel-row-best-bound-is-used-for-rhs", "[presolve]" )
 {
    Num<double> num{};
    double time = 0.0;
+int cause = -1;
    Timer t{ time };
    Message msg{};
    Problem<double> problem =
@@ -811,10 +869,10 @@ TEST_CASE( "parallel-row-best-bound-is-used-for-rhs", "[presolve]" )
    Reductions<double> reductions{};
 
    PresolveStatus presolveStatus =
-       presolvingMethod.execute( problem, problemUpdate, num, reductions, t );
+       presolvingMethod.execute( problem, problemUpdate, num, reductions, t, cause );
 
    REQUIRE( presolveStatus == PresolveStatus::kReduced );
-   REQUIRE( reductions.size() == 7 );
+   REQUIRE( reductions.size() == 8 );
    int locked_rows[] = { 0, 1, 2 };
    for( int i = 0; i < 3; i++ )
    {
@@ -823,20 +881,20 @@ TEST_CASE( "parallel-row-best-bound-is-used-for-rhs", "[presolve]" )
       REQUIRE( reductions.getReduction( i ).newval == 0 );
    }
 
-   REQUIRE( reductions.getReduction( 3 ).col ==
+   REQUIRE( reductions.getReduction( 4 ).col ==
             RowReduction::REASON_FOR_LESS_RESTRICTIVE_BOUND_CHANGE );
 
-   REQUIRE( reductions.getReduction( 4 ).col ==
+   REQUIRE( reductions.getReduction( 5 ).col ==
             RowReduction::LHS_LESS_RESTRICTIVE );
-   REQUIRE( reductions.getReduction( 4 ).row == 0 );
-   REQUIRE( reductions.getReduction( 4 ).newval == -1 );
+   REQUIRE( reductions.getReduction( 5 ).row == 0 );
+   REQUIRE( reductions.getReduction( 5 ).newval == -1 );
 
    for( int i = 0; i < 2; i++ )
    {
-      REQUIRE( reductions.getReduction( 5 + i ).col ==
+      REQUIRE( reductions.getReduction( 6 + i ).col ==
                RowReduction::REDUNDANT );
-      REQUIRE( reductions.getReduction( 5 + i ).row == locked_rows[i + 1] );
-      REQUIRE( reductions.getReduction( 5 + i ).newval == 0 );
+      REQUIRE( reductions.getReduction( 6 + i ).row == locked_rows[i + 1] );
+      REQUIRE( reductions.getReduction( 6 + i ).newval == 0 );
    }
 }
 
@@ -844,6 +902,7 @@ TEST_CASE( "parallel-row-best-bound-is-used-for-rhs-coeff-not-1", "[presolve]" )
 {
    Num<double> num{};
    double time = 0.0;
+int cause = -1;
    Timer t{ time };
    Message msg{};
    Problem<double> problem =
@@ -859,10 +918,10 @@ TEST_CASE( "parallel-row-best-bound-is-used-for-rhs-coeff-not-1", "[presolve]" )
    Reductions<double> reductions{};
 
    PresolveStatus presolveStatus =
-       presolvingMethod.execute( problem, problemUpdate, num, reductions, t );
+       presolvingMethod.execute( problem, problemUpdate, num, reductions, t, cause );
 
    REQUIRE( presolveStatus == PresolveStatus::kReduced );
-   REQUIRE( reductions.size() == 7 );
+   REQUIRE( reductions.size() == 8 );
    int locked_rows[] = { 0, 1, 2 };
    for( int i = 0; i < 3; i++ )
    {
@@ -871,20 +930,20 @@ TEST_CASE( "parallel-row-best-bound-is-used-for-rhs-coeff-not-1", "[presolve]" )
       REQUIRE( reductions.getReduction( i ).newval == 0 );
    }
 
-   REQUIRE( reductions.getReduction( 3 ).col ==
+   REQUIRE( reductions.getReduction( 4 ).col ==
             RowReduction::REASON_FOR_LESS_RESTRICTIVE_BOUND_CHANGE );
 
-   REQUIRE( reductions.getReduction( 4 ).col ==
+   REQUIRE( reductions.getReduction( 5 ).col ==
             RowReduction::LHS_LESS_RESTRICTIVE );
-   REQUIRE( reductions.getReduction( 4 ).row == 0 );
-   REQUIRE( reductions.getReduction( 4 ).newval == -1 );
+   REQUIRE( reductions.getReduction( 5 ).row == 0 );
+   REQUIRE( reductions.getReduction( 5 ).newval == -1 );
 
    for( int i = 0; i < 2; i++ )
    {
-      REQUIRE( reductions.getReduction( 5 + i ).col ==
+      REQUIRE( reductions.getReduction( 6 + i ).col ==
                RowReduction::REDUNDANT );
-      REQUIRE( reductions.getReduction( 5 + i ).row == locked_rows[i + 1] );
-      REQUIRE( reductions.getReduction( 5 + i ).newval == 0 );
+      REQUIRE( reductions.getReduction( 6 + i ).row == locked_rows[i + 1] );
+      REQUIRE( reductions.getReduction( 6 + i ).newval == 0 );
    }
 }
 
@@ -892,6 +951,7 @@ TEST_CASE( "parallel-row-multiple-parallel-rows", "[presolve]" )
 {
    Num<double> num{};
    double time = 0.0;
+int cause = -1;
    Timer t{ time };
    Message msg{};
    Problem<double> problem = setupParallelRowWithMultipleParallelRows();
@@ -906,10 +966,10 @@ TEST_CASE( "parallel-row-multiple-parallel-rows", "[presolve]" )
    Reductions<double> reductions{};
 
    PresolveStatus presolveStatus =
-       presolvingMethod.execute( problem, problemUpdate, num, reductions, t );
+       presolvingMethod.execute( problem, problemUpdate, num, reductions, t, cause );
 
    REQUIRE( presolveStatus == PresolveStatus::kReduced );
-   REQUIRE( reductions.size() == 7 );
+   REQUIRE( reductions.size() == 8 );
    int locked_rows[] = { 1, 2, 0, 3 };
    for( int i = 0; i < 4; i++ )
    {
@@ -920,10 +980,10 @@ TEST_CASE( "parallel-row-multiple-parallel-rows", "[presolve]" )
 
    for( int i = 0; i < 3; i++ )
    {
-      REQUIRE( reductions.getReduction( 4 + i ).col ==
+      REQUIRE( reductions.getReduction( 5 + i ).col ==
                RowReduction::REDUNDANT );
-      REQUIRE( reductions.getReduction( 4 + i ).row == locked_rows[i + 1] );
-      REQUIRE( reductions.getReduction( 4 + i ).newval == 0 );
+      REQUIRE( reductions.getReduction( 5 + i ).row == locked_rows[i + 1] );
+      REQUIRE( reductions.getReduction( 5 + i ).newval == 0 );
    }
 }
 
@@ -1214,7 +1274,7 @@ setupProblemParallelRowWithMixed( bool firstRowEquation, double lhsIneq,
    pb.setColNameAll( columnNames );
    pb.setProblemName( "matrix with parallel inequalities (0 and 2)" );
    Problem<double> problem = pb.build();
-   int i = firstRowEquation ? i = 0 : i = 2;
+   int i = firstRowEquation ? 0 : 2;
    problem.getConstraintMatrix().modifyLeftHandSide( i, num, rhs[i] );
    return problem;
 }
